@@ -41,6 +41,12 @@ struct
   int symmetric;
   const char *filename;
   const char *directory;
+  int batch;
+  int answer_yes;
+  int answer_no;
+  int status_fd;
+  int require_compliance;
+  int with_log;
 } opt;
 
 
